local KeywordColor = "e8a038";
local KeywordSound = "Sound\\interface\\AuctionWindowOpen.wav";
local ChatFont = "Interface\\AddOns\\ChatAssist\\font01.ttf";
local URLColor = "0000ff";
local ClassColor = {
 "ff7c0a", -- Druid
 "aad372", -- Hunter
 "68cbef", -- Mage
 "f48cba", -- Paladin
 "ffffff", -- Priest
 "fff468", -- Rouge
 "00dbba", -- Shaman
 "9382c9", -- Warlock
 "c69b6d", -- Warrior
};
-- Please do not change below this line.
local Version = '2.00.20003';

local DefaultProfile = {
    Keywords = {"Keyword"},
    ChannelSticky = true,
    OfficerSticky = false,
    WhisperSticky = false,
    Timestamp = true,
    TimestampFormat = "%H:%M",
    Logging = false,
    Window = { true, false, true, true, true, true, true },
    Locale = "enUS",
    ColorName = true,
    ShortTag = false,
    HideTag = false,
    HideChannel = false,
    UnicodeBlock = { enable = true, guild = false, party = false, raid = false, raidwarning = false, officer = false },
    DefaultChat = "say",
    OnScreenFontSize = 24,
    Event = {
        KEYWORD = { OnScreen = false, OnScreenColor = { r = 0.85, g = 0.37, b = 0.00 }},
        AUCTION = { OnScreen = false, OnScreenColor = { r = 0.76, g = 0.00, b = 0.40 }},
        RAIDLEADER = { OnScreen = false, OnScreenColor = { r = 0.99, g = 0.98, b = 0.00 }},
        BGLEADER = { OnScreen = false, OnScreenColor = { r = 0.99, g = 0.98, b = 0.00 }},
        OFFICER = { OnScreen = false, OnScreenColor = { r = 0.35, g = 0.67, b = 0.42 }},
        WHISPER = { OnScreen = false, OnScreenColor = { r = 0.87, g = 0.41, b = 0.58 }},
    },
};
local KeywordLog = {};
local AuctionLog = {};
local ClassCache = {};
local KeywordInterval = 2;
local LastKeyword = nil;
local LastKeywordTime = 0;

ChatAssist = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceHook-2.1", "AceDB-2.0");
local L = AceLibrary("AceLocale-2.2"):new("ChatAssist");

function ChatAssist:OnInitialize()
    self:RegisterEvent("PLAYER_ENTERING_WORLD");
    self:RegisterEvent("PLAYER_LEAVING_WORLD");
    self:RegisterEvent("ADDON_LOADED");
    self:RegisterEvent("VARIABLES_LOADED");
    self:RegisterEvent("PARTY_MEMBERS_CHANGED");
    self:RegisterEvent("RAID_ROSTER_UPDATE");
    self:RegisterEvent("GUILD_ROSTER_UPDATE");
    self:RegisterEvent("FRIENDLIST_UPDATE");
    self:RegisterEvent("CHAT_MSG_RAID_LEADER");
    self:RegisterEvent("CHAT_MSG_BATTLEGROUND_LEADER");
    self:RegisterEvent("CHAT_MSG_OFFICER");
    self:RegisterEvent("CHAT_MSG_WHISPER");

    self:RegisterDB("ChatAssistDB", "ChatAssistDBPC");
    self:RegisterDefaults("profile", DefaultProfile);
    local command_options = {
        type = 'group',
        args = {
            add = {
                type = 'text',
                name = 'Add Keyword',
                desc = 'Add Keyword',
                usage = '<Keyword>',
                get = false,
                set = 'Command_addKeyword',
            },
            del = {
                type = 'text',
                name = 'Delete Keyword',
                desc = 'Delete Keyword',
                usage = '<Keyword>',
                get = false,
                set = 'Command_delKeyword',
            },
            list = {
                type = 'execute',
                name = 'List Keyword',
                desc = 'List Keyword',
                func = 'Command_listKeyword',
            },
            channelsticky = {
                type = 'toggle',
                name = 'Toggle ChannelSticky',
                desc = 'Toggle ChannelSticky',
                get = 'Command_showChannelSticky',
                set = 'Command_toggleChannelSticky',
            },
            officersticky = {
                type = 'toggle',
                name = 'Toggle OfficerSticky',
                desc = 'Toggle OfficerSticky',
                get = 'Command_showOfficerSticky',
                set = 'Command_toggleOfficerSticky',
            },
            whispersticky = {
                type = 'toggle',
                name = 'Toggle WhisperSticky',
                desc = 'Toggle WhisperSticky',
                get = 'Command_showWhisperSticky',
                set = 'Command_toggleWhisperSticky',
            },
            ts = {
                type = 'toggle',
                name = 'Toggle TimeStamp',
                desc = 'Toggle TimeStamp',
                get = 'Command_showTimestamp',
                set = 'Command_toggleTimestamp',
            },
            tsformat = {
                type = 'text',
                name = 'Set TimeStamp Format',
                desc = 'Set TimeStamp Format',
                usage = '<TimeStampFormat>',
                get = 'Command_getTimestampFormat',
                set = 'Command_setTimestampFormat',
            },
            window = {
                type = 'text',
                name = 'Toggle Window',
                desc = 'Toggle Window',
                usage = '<WindowId>',
                validate = { '1', '2', '3', '4', '5', '6', '7' },
                get = false,
                set = 'Command_toggleWindow',
            },
            log = {
                type = 'toggle',
                name = 'Toggle Logging',
                desc = 'Toggle Logging',
                get = 'Command_showLog',
                set = 'Command_toggleLog',
            },
            history = {
                type = 'execute',
                name = 'Show Keyword/Auction History',
                desc = 'Show Keyword/Auction History',
                func = 'Command_History',
            },
            colorname = {
                type = 'toggle',
                name = 'Toggle ColorName',
                desc = 'Toggle ColorName',
                get = 'Command_showColorName',
                set = 'Command_toggleColorName',
            },
            shorttag = {
                type = 'toggle',
                name = 'Toggle ShortTag',
                desc = 'Toggle ShortTag',
                get = 'Command_showShortTag',
                set = 'Command_toggleShortTag',
            },
            hidetag = {
                type = 'toggle',
                name = 'Toggle HideTag',
                desc = 'Toggle HideTag',
                get = 'Command_showHideTag',
                set = 'Command_toggleHideTag',
            },
            hidechannel = {
                type = 'toggle',
                name = 'Toggle HideChannel',
                desc = 'Toggle HideChannel',
                get = 'Command_showHideChannel',
                set = 'Command_toggleHideChannel',
            },
            ub = {
                type = 'text',
                name = 'Toggle UnicodeBlock',
                desc = 'Toggle UnicodeBlock',
                usage = '<enable|party|raid|raidwarning|guild|officer>',
                validate = { 'enable', 'party', 'raid', 'raidwarning', 'guild', 'officer' },
                get = 'Command_getUnicodeBlock',
                set = 'Command_toggleUnicodeBlock',
            },
            defaultchat = {
                type = 'text',
                name = 'Set DefaultChat',
                desc = 'Set DefaultChat Mode',
                usage = '<say|party|raid|guild|officer>',
                validate = { 'say', 'party', 'raid', 'guild', 'officer' },
                get = 'Command_getDefaultChat',
                set = 'Command_setDefaultChat',
            },
            onscreen = {
                type = 'text',
                name = 'On-screen displays',
                desc = 'On-screen displays',
                usage = '<keyword|auction|raidleader|bgleader|officer|whisper>',
                validate = { 'keyword', 'auction', 'raidleader', 'bgleader', 'officer', 'whipser' },
                get = false,
                set = 'Command_toggleOnScreen',
            },
            ccrebuild = {
                type = 'execute',
                name = 'Rebuild ClassCache',
                desc = 'Rebuild ClassCache',
                func = 'Command_CCRebuild',
            },
            menu = {
                type = 'execute',
                name = 'Show Options',
                desc = 'Show Options',
                func = 'Command_Menu',
            },
            frame = {
                type = 'execute',
                name = 'Walk Frames',
                desc = 'Walk Frames',
                func = 'Command_Frame',
            },
        },
    };
    local tt_options = {
        type = 'text',
        name = 'TellTarget',
        desc = 'Tell to Target',
        usage = '<Message>',
        get = false,
        set = 'Command_TellTarget',
    };
    self:RegisterChatCommand({'/chatassist', '/ca'}, command_options);
    self:RegisterChatCommand({'/tt'}, tt_options);

    -- register for myAddOns
    if(myAddOnsFrame_Register ~= nil) then
      local info = {
        name = L['ChatAssist'];
        version = Version;
        author = "ayu";
        website = "http://fed.commun.jp/wow/";
        category = L['Chat'];
        optionsframe = "ChatAssistOptionsFrame";
      };
      myAddOnsFrame_Register(info);
    end
end

function ChatAssist:OnEnable()
    for i = 1, 7 do
        local func = function(...)
            return self:AddMessage(i, ...);
        end
        self:Hook(getglobal('ChatFrame' .. i), "AddMessage", func, true);
    end
    self:Hook('SetItemRef', true);
    self:Hook('SendChatMessage', true);
end

function ChatAssist:OnDisable()
    self:UnhookAll();
end

--
-- Event Callbacks
--

function ChatAssist:PLAYER_ENTERING_WORLD()
    self:RegisterEvent("PARTY_MEMBERS_CHANGED");
    self:RegisterEvent("RAID_ROSTER_UPDATE");
    self:RegisterEvent("GUILD_ROSTER_UPDATE");
    self:RegisterEvent("FRIENDLIST_UPDATE");
    self:RegisterEvent("CHAT_MSG_RAID_LEADER");
    self:RegisterEvent("CHAT_MSG_BATTLEGROUND_LEADER");
    self:RegisterEvent("CHAT_MSG_OFFICER");
    self:RegisterEvent("CHAT_MSG_WHISPER");

    if(GetNumRaidMembers() > 0) then
        ChatAssist:UpdateClassCache("RAID");
    end

    if(self.db.profile.ColorName) then
        -- getPlayerName and getPlayerClass
        ChatAssist:UpdateClassCache("PLAYER");

        if(IsInGuild()) then
            GuildRoster();
        end
    end
end

function ChatAssist:PLAYER_LEAVING_WORLD()
    self:UnregisterEvent("PARTY_MEMBERS_CHANGED");
    self:UnregisterEvent("RAID_ROSTER_UPDATE");
    self:UnregisterEvent("GUILD_ROSTER_UPDATE");
    self:UnregisterEvent("FRIENDLIST_UPDATE");
    self:UnregisterEvent("CHAT_MSG_RAID_LEADER");
    self:UnregisterEvent("CHAT_MSG_BATTLEGROUND_LEADER");
    self:UnregisterEvent("CHAT_MSG_OFFICER");
    self:UnregisterEvent("CHAT_MSG_WHISPER");
end

function ChatAssist:ADDON_LOADED()
    -- Replace Font
    ChatFontNormal:SetFont(ChatFont, 14);

    -- RaidWarning
    RaidWarningFrame:SetFont(ChatFont, 18);

    -- Replace Mail Font
    --SendMailSubjectEditBox:SetFont(ChatFont, 14);
    SendMailBodyEditBox:SetFont(ChatFont, 14);
    for i=1, INBOXITEMS_TO_DISPLAY do
      getglobal("MailItem"..i.."Subject"):SetFont(ChatFont, 14);
    end
    OpenMailSubject:SetFont(ChatFont, 14);
    OpenMailBodyText:SetFont(ChatFont, 14);

    -- Replace Social/Guild 
    GuildFrameNotesText:SetFont(ChatFont, 11);
    GuildInfoEditBox:SetFont(ChatFont, 11);
    PersonalNoteText:SetFont(ChatFont, 10);
    OfficerNoteText:SetFont(ChatFont, 10);

    -- Replace CT_RaidAssist (CT_RAMessageFrame) Font
    if(CT_RAMessageFrame ~= nil) then
        CT_RAMessageFrame:SetFont(ChatFont, 18);
    end
end

function ChatAssist:VARIABLES_LOADED()
    -- MessageFont
    if(ChatAssist_MessageFont:SetFont(ChatFont, self.db.profile.OnScreenFontSize) == nil) then
        ChatAssist_MessageFont:SetFont("Fonts\\FRIZQT__.TTF", self.db.profile.OnScreenFontSize);
    end

    -- Logging
    if(self.db.profile.Logging) then
      LoggingChat(true);
    end

    -- Channel Sticky
    if(self.db.profile.ChannelSticky) then
      ChatTypeInfo["CHANNEL"].sticky = 1;
    end

    -- Officer Sticky
    if(self.db.profile.OfficerSticky) then
      ChatTypeInfo["OFFICER"].sticky = 1;
    end

    -- Whisper Sticky
    if(self.db.profile.WhisperSticky) then
      ChatTypeInfo["WHISPER"].sticky = 1;
    end

    -- ChatType
    if(self.db.profile.DefaultChat) then
      --DEFAULT_CHAT_FRAME.editBox.chatType = string.upper(self.db.profile.DefaultChat);
      --DEFAULT_CHAT_FRAME.editBox.stickyType = string.upper(self.db.profile.DefaultChat);
      DEFAULT_CHAT_FRAME.editBox:SetAttribute("chatType", string.upper(self.db.profile.DefaultChat));
      DEFAULT_CHAT_FRAME.editBox:SetAttribute("stickyType", string.upper(self.db.profile.DefaultChat));
    end

    -- Variable Load Complete
    DEFAULT_CHAT_FRAME:AddMessage(string.format(L["ChatAssist Ver %s Loaded. |c0000ff00(/ca for options)"], Version));
end

function ChatAssist:PARTY_MEMBERS_CHANGED()
    ChatAssist:UpdateClassCache("PARTY");
end

function ChatAssist:RAID_ROSTER_UPDATE()
    ChatAssist:UpdateClassCache("RAID");
end

function ChatAssist:GUILD_ROSTER_UPDATE()
    ChatAssist:UpdateClassCache("GUILD");
end

function ChatAssist:FRIENDLIST_UPDATE()
    ChatAssist:UpdateClassCache("FRIEND");
end

function ChatAssist:CHAT_MSG_RAID_LEADER()
    if(self.db.profile.Event.RAIDLEADER.OnScreen) then
        ChatAssist_MessageFrame:AddMessage("[" .. arg2 .. "]: " .. arg1, self.db.profile.Event.RAIDLEADER.OnScreenColor.r, self.db.profile.Event.RAIDLEADER.OnScreenColor.g, self.db.profile.Event.RAIDLEADER.OnScreenColor.b, 1.0, UIERRORS_HOLD_TIME);
    end
end

function ChatAssist:CHAT_MSG_BATTLEGROUND_LEADER()
    if(self.db.profile.Event.BGLEADER.OnScreen == 1) then
        ChatAssist_MessageFrame:AddMessage("[" .. arg2 .. "]: " .. arg1, self.db.profile.Event.BGLEADER.OnScreenColor.r, self.db.profile.Event.BGLEADER.OnScreenColor.g, self.db.profile.Event.BGLEADER.OnScreenColor.b, 1.0, UIERRORS_HOLD_TIME);
    end
end

function ChatAssist:CHAT_MSG_OFFICER()
    if(self.db.profile.Event.OFFICER.OnScreen == 1) then
        ChatAssist_MessageFrame:AddMessage("[" .. arg2 .. "]: " .. arg1, self.db.profile.Event.OFFICER.OnScreenColor.r, self.db.profile.Event.OFFICER.OnScreenColor.g, self.db.profile.Event.OFFICER.OnScreenColor.b, 1.0, UIERRORS_HOLD_TIME);
    end
end

function ChatAssist:CHAT_MSG_WHISPER()
    if(self.db.profile.Event.WHISPER.OnScreen == 1) then
        ChatAssist_MessageFrame:AddMessage("[" .. arg2 .. "]whispers: " .. arg1, self.db.profile.Event.WHISPER.OnScreenColor.r, self.db.profile.Event.WHISPER.OnScreenColor.g, self.db.profile.Event.WHISPER.OnScreenColor.b, 1.0, UIERRORS_HOLD_TIME);
    end
end


--
-- Hook Functions
--

function ChatAssist:AddMessage(id, f, t, ...)
    if (self.db.profile.Window[id]) then
        t = self:ParseMessage(t);
    end
    return self.hooks[f].AddMessage(f, t, ...);
end

function ChatAssist:SetItemRef(link, text, button)
  local type = strsub(link, 1, 3);
  if(type == "url") then
    ChatAssist_URLFrame:Show();
    ChatAssist_URLFrameURLEditBox:SetText(strsub(link, 5));
  else
    self.hooks["SetItemRef"](link, text, button);
  end
end

function ChatAssist:SendChatMessage(msg, ...)
    local channel_name;
    local system, language, channel_id = select(1, ...);

    if ((msg == nil or msg == "") and system ~= "AFK" and system ~= "DND") then
        return;
    end

    -- Channel
    if (system == "CHANNEL") then
        channel_id, channel_name = GetChannelName(channel_id);
        if (channel_id == 0 or channel_name == nil) then
            return;
        end

        if (not string.find(channel_name, "^".. L["General"] .." -") and not string.find(channel_name, "^".. L["Trade"] .." -") and not string.find(channel_name, "^".. L["LocalDefense"] .. " -") and not string.find(channel_name, "^".. L["WorldDefense"] .." -") and not string.find(channel_name, "^".. L["LookingForGroup"] .." -") and not string.find(channel_name, "^".. L["GuildRecruitment"])) then
            msg = string.gsub(msg, "|c(%x+)|Hitem:([0-9]+):([0-9]+):([0-9]+):([0-9]+)|h%[([^%]]+)%]|h|r", "{CLINK:%1:%2:%3:%4:%5:%6}");
        end
    end

    if (ChatAssist:isMultiByte(msg) and self.db.profile.UnicodeBlock["enable"] == 1) then
        -- Say Emote Shout check
        if (system == "SAY" or system == "EMOTE" or system == "YELL" or system == "BATTLEGROUND") then
            self:Print(L["Unicode cannot be used with public channel. |c0000ff00(type \"/ca ub\" to disable)"]);
            return;
        elseif (system == "CHANNEL") then
            if (string.find(channel_name, "^".. L["General"] .." -") or string.find(channel_name, "^".. L["Trade"] .." -") or string.find(channel_name, "^".. L["LocalDefense"] .. " -") or string.find(channel_name, "^".. L["WorldDefense"] .." -") or string.find(channel_name, "^".. L["LookingForGroup"] .." -") or string.find(channel_name, "^".. L["GuildRecruitment"])) then
                self:Print(L["Unicode cannot be used with public channel. |c0000ff00(type \"/ca ub\" to disable)"]);
                return;
            end
        elseif (system == "PARTY" and self.db.profile.UnicodeBlock["party"]) then
            self:Print(L["Unicode cannot be used with public channel. |c0000ff00(type \"/ca ub\" to disable)"]);
            return;
        elseif(system == "RAID" and self.db.profile.UnicodeBlock["raid"]) then
            self:Print(L["Unicode cannot be used with public channel. |c0000ff00(type \"/ca ub\" to disable)"]);
            return;
        elseif(system == "RAID_WARNING" and self.db.profile.UnicodeBlock["raidwarning"]) then
            self:Print(L["Unicode cannot be used with public channel. |c0000ff00(type \"/ca ub\" to disable)"]);
            return;
        elseif(system == "GUILD" and self.db.profile.UnicodeBlock["guild"]) then
            self:Print(L["Unicode cannot be used with public channel. |c0000ff00(type \"/ca ub\" to disable)"]);
            return;
        elseif(system == "OFFICER" and self.db.profile.UnicodeBlock["officer"]) then
            self:Print(L["Unicode cannot be used with public channel. |c0000ff00(type \"/ca ub\" to disable)"]);
            return;
        end
        self.hooks["SendChatMessage"](msg, ...);
    else
        self.hooks["SendChatMessage"](msg,...);
    end
end

-- MouseWheel Enabled
function ChatAssist:OnMouseWheel(frame, value)
    if (IsShiftKeyDown()) then
        if (value > 0) then
            frame:ScrollToTop();
        else
            frame:ScrollToBottom();
        end
    elseif (IsControlKeyDown()) then
        if (value > 0) then
            frame:PageUp();
        else
            frame:PageDown();
        end
    else
        if (value > 0) then
            frame:ScrollUp();
        else
            frame:ScrollDown();
        end
    end
end

--
-- Internal Functions
--

function ChatAssist:ParseMessage(text)
    if (text == nil) then
        return text;
    end

    -- ShortTag
    if (self.db.profile.ShortTag and string.find(text, ":%s") and (string.find(text, "|Hplayer:") or string.find(text, "%s:%s"))) then
        text = string.gsub(text, "%[%d+%.%s" .. L["General"] .. "%]", "[" .. L["GE"] .. "]");
        text = string.gsub(text, "%[%d+%.%s" .. L["Trade"] .. "%]", "[" .. L["TR"] .. "]");
        text = string.gsub(text, "%[%d+%.%s" .. L["LookingForGroup"] .. "%]", "[" .. L["LFG"] .. "]");
        text = string.gsub(text, "%[%d+%.%s" .. L["WorldDefense"] .. "%]", "[" .. L["WD"] .. "]");
        text = string.gsub(text, "%[%d+%.%s" .. L["LocalDefense"] .. "%]", "[" .. L["LD"] .. "]");
        text = string.gsub(text, "%[%d+%.%s" .. L["GuildRecruitment"] .. "%]", "[" .. L["GR"] .. "]");
        text = string.gsub(text, "%[" .. L["Party"] .. "%]", "[".. L["PTY"] .. "]");
        text = string.gsub(text, "%[" .. L["Raid"] .. "%]", "[".. L["RD"] .. "]");
        text = string.gsub(text, "%[" .. L["Raid Leader"] .. "%]", "[".. L["RL"] .. "]");
        text = string.gsub(text, "%[" .. L["Raid Warning"] .. "%]", "[".. L["RW"] .. "]");
        text = string.gsub(text, "%[" .. L["Battleground"] .. "%]", "[".. L["BG"] .. "]");
        text = string.gsub(text, "%[" .. L["Battleground Leader"] .. "%]", "[".. L["BGL"] .. "]");
        text = string.gsub(text, "%[" .. L["Guild"] .. "%]", "[".. L["GL"] .. "]");
        text = string.gsub(text, "%[" .. L["Officer"] .. "%]", "[".. L["OF"] .. "]");
        text = string.gsub(text, "%[" .. "(%d+)%.%s%w+" .. "%]", "[%1]");
    -- HideTag
    elseif (self.db.profile.HideTag and string.find(text, ":%s") and (string.find(text, "|Hplayer:") or string.find(text, "%s:%s"))) then
        text = string.gsub(text, "%[" .. L["Party"] .. "%]%s", "");
        text = string.gsub(text, "%[" .. L["Raid"] .. "%]%s", "");
        text = string.gsub(text, "%[" .. L["Raid Leader"] .. "%]%s", "");
        text = string.gsub(text, "%[" .. L["Raid Warning"] .. "%]%s", "");
        text = string.gsub(text, "%[" .. L["Battleground"] .. "%]%s", "");
        text = string.gsub(text, "%[" .. L["Battleground Leader"] .. "%]%s", "");
        text = string.gsub(text, "%[" .. L["Guild"] .. "%]%s", "");
        text = string.gsub(text, "%[" .. L["Officer"] .. "%]%s", "");

        if (self.db.profile.HideChannel) then
            text = string.gsub(text, "%[" .. "(%d+)%.%s%w+" .. "%]%s", "");
            if(string.sub(text, 1, 2) == ": ") then
                text = string.sub(text, 3);
            end
        else
            text = string.gsub(text, "%[" .. "(%d+)%.%s%w+" .. "%]", "[%1]");
        end
    end

    -- ColorName
    if (self.db.profile.ColorName and string.find(text, "|Hplayer:")) then
        local name = string.gsub(text, ".*|Hplayer:([%w%-]-)|h%[([%w%-]-)%]|h.*", "%1");
        local mist = string.find(name, "-", 1, true);
        local name1, class = 0;
        if (mist ~= nil) then
            name1 = string.sub(name, 1, mist - 1);
            class = ChatAssist:GetClassCache(name1);
        else
            class = ChatAssist:GetClassCache(name);
        end

        if (class ~= nil and class ~= 0) then
            name = string.gsub(name, "%-", "%%-");
            text = string.gsub(text, "|Hplayer:".. name .. "|h%[".. name .. "%]|h", "[|cff" .. ClassColor[class] .. "|Hplayer:" .. name .. "|h" .. name .. "|h|r]");
        end
    end

    -- URL
    local urlreplace = 0;
    text, urlreplace = string.gsub(text, "(https?://[$-_%.!~%*'%(%)a-zA-Z0-9;/%?:@&=%+%$,%%#]+)", "|cff" .. URLColor .. "|Hurl:%1|h[%1]|h|r");
    if (urlreplace == 0) then
        text = string.gsub(text, "(www%.[$-_%.!~%*'%(%)a-zA-Z0-9;/%?:@&=%+%$,%%#]+)", "|cff" .. URLColor .. "|Hurl:%1|h[%1]|h|r");
    end

    -- ItemLink
    text = string.gsub(text, "{CLINK:(%x+):([0-9]+):([0-9]+):([0-9]+):([0-9]+):([^}]+)}", "|c%1|Hitem:%2:%3:%4:%5|h[%6]|h|r");

    -- Timestamp
    if (self.db.profile.Timestamp) then
        local ts = self.db.profile.TimestampFormat;
        local date_H = date("%H");
        local date_M = date("%M");
        local date_S = date("%S");
        local date_I = date("%I");
        local date_P = "";
        local date_p = "";

        if (tonumber(date_H) < 12) then
            date_P = "AM";
            date_p = "am";
        else
            date_P = "PM";
            date_p = "pm";
        end

        ts = string.gsub(ts, "%%H", date_H);
        ts = string.gsub(ts, "%%M", date_M);
        ts = string.gsub(ts, "%%S", date_S);
        ts = string.gsub(ts, "%%I", date_I);
        ts = string.gsub(ts, "%%P", date_P);
        ts = string.gsub(ts, "%%p", date_p);
        text = "[" .. ts .. "] " .. text;
    end

    -- AuctionLog
    for key,value in pairs(self.AuctionMessages) do
        if (string.find(text, value)) then
            -- AuctionLog
            table.insert(AuctionLog, text);
            if (table.getn(AuctionLog) > 30) then
                table.remove(AuctionLog, 1);
            end
            ChatAssist:HistoryFrameChange("AUCTION");

            if (self.db.profile.Event.AUCTION.OnScreen == 1) then
                ChatAssist_MessageFrame:AddMessage(text, self.db.profile.Event.AUCTION.OnScreenColor.r, self.db.profile.Event.AUCTION.OnScreenColor.g, self.db.profile.Event.AUCTION.OnScreenColor.b, 1.0, UIERRORS_HOLD_TIME);
            end
            break;
        end
    end

    -- Keyword
    local st, ed = string.find(text, ": ");
    if(ed ~= nil and string.find(text, "|Hplayer"))then
        local tag = string.sub(text, 1, st -1);
        local chat = string.sub(text, ed + 1, -1);
        local match = 0;
        for key,value in pairs(self.db.profile.Keywords) do
            if(string.find(chat, value) ~= nil) then
                chat = string.gsub(chat, value, "|cff" .. KeywordColor .. value .. "|r");
                match = 1;
            end
        end
        if (match == 1) then
            text = tag .. ": " .. chat;
            if (LastKeyword ~= text or LastKeywordTime < time() - KeywordInterval) then
                LastKeyword = text;
                LastKeywordTime = time();

                -- KeywordLog
                table.insert(KeywordLog, text);
                if (table.getn(KeywordLog) > 30) then
                    table.remove(KeywordLog, 1);
                end
                ChatAssist:HistoryFrameChange("KEYWORD");

                -- PlaySound
                if (KeywordSound ~= nil) then
                    PlaySoundFile(KeywordSound);
                end

                if (self.db.profile.Event.KEYWORD.OnScreen) then
                    ChatAssist_MessageFrame:AddMessage(text, self.db.profile.Event.KEYWORD.OnScreenColor.r, self.db.profile.Event.KEYWORD.OnScreenColor.g, self.db.profile.Event.KEYWORD.OnScreenColor.b, 1.0, UIERRORS_HOLD_TIME);
                end
            end
        end
    end

    return text;
end

function ChatAssist:GetClassCache(name)
    if(ClassCache[name] ~= nil) then
        return ClassCache[name];
    end
    return 0;
end

function ChatAssist:SetClassCache(name, class)
    if(name == nil or class == nil or class == 0) then
        return;
    end
    ClassCache[name] = class;
end

function ChatAssist:UpdateClassCache(type)
    if (type == "PLAYER") then
        local class, eng_class = UnitClass("player");
        local classid = ChatAssist:ClassToNumber(eng_class);
        if(classid ~= 0) then
            ChatAssist:SetClassCache(UnitName("player"), classid);
        end
    elseif (type == "PARTY") then
        for i = 1, GetNumPartyMembers() do
            local class, eng_class = UnitClass("party" .. i);
            local classid = ChatAssist:ClassToNumber(eng_class);
            if (classid ~= 0) then
                ChatAssist:SetClassCache(UnitName("party" .. i), classid);
            end
        end
    elseif (type == "RAID") then
        for i = 1, GetNumRaidMembers() do
            local name, rank, subgroup, level, class, filename, zone, online = GetRaidRosterInfo(i);
            local classid = ChatAssist:ClassToNumber(filename);
            if (classid ~= 0) then
                ChatAssist:SetClassCache(name, classid);
            end
       end
    elseif (type == "GUILD") then
        for i = 1, GetNumGuildMembers(true) do
            local name, rank, rankIndex, level, class, zone, note, officernote, online, status = GetGuildRosterInfo(i);
            local classid = ChatAssist:ClassToNumber(class);
            if (classid ~= 0) then
                ChatAssist:SetClassCache(name, classid);
            end
        end
    elseif (type == "FRIEND") then
        for i = 1, GetNumFriends() do
            local name, level, class, area, connected, status = GetFriendInfo(i);
            local classid = ChatAssist:ClassToNumber(class);
            if (classid ~= 0) then
                ChatAssist:SetClassCache(name, classid);
            end
        end
    end
end

function ChatAssist_RebuildClassCache()
    ChatAssist:ClearClassCache();
    ChatAssist:UpdateClassCache("PLAYER");
    ChatAssist:UpdateClassCache("PARTY");
    ChatAssist:UpdateClassCache("RAID");
    ChatAssist:UpdateClassCache("FRIEND");
    if (IsInGuild()) then
        GuildRoster();
    end
end

function ChatAssist:ClearClassCache()
    ClassCache = {};
end

function ChatAssist:GCClassCache(type)
end

function ChatAssist:ClassToNumber(class)
    if (class == nil) then
        return 0;
    end
    class = string.upper(class);
    if (class == "DRUID") then
        return 1;
    elseif (class == "HUNTER") then
        return 2;
    elseif (class == "MAGE") then
        return 3;
    elseif (class == "PALADIN") then
        return 4;
    elseif (class == "PRIEST") then
        return 5;
    elseif (class == "ROGUE") then
        return 6;
    elseif (class == "SHAMAN") then
        return 7;
    elseif (class == "WARLOCK") then
        return 8;
    elseif (class == "WARRIOR") then
        return 9;
    else
        return 0;
    end
end

function ChatAssist:HistoryFrameToggle(tab)
    if (ChatAssist_HistoryFrame:IsVisible()) then
        ChatAssist_HistoryFrame:Hide();
    else
        ChatAssist_HistoryFrame:Show();
        ChatAssist:HistoryFrameChange(tab);
    end
end

function ChatAssist:HistoryFrameChange(tab)
    if (ChatAssist_HistoryFrame:IsVisible()) then
        if (tab == "KEYWORD") then
            ChatAssist_HistoryFrameScrollFrameString:SetText(ChatAssist:strjoin("\n", KeywordLog));
            ChatAssist_HistoryFrameScrollFrame:SetHeight(ChatAssist_HistoryFrameScrollFrameString:GetHeight()+10);
            ChatAssist_HistoryFrameScroll:UpdateScrollChildRect();
            ChatAssist_HistoryFrameKeywordButton:Disable();
            ChatAssist_HistoryFrameAuctionButton:Enable();
        elseif (tab == "AUCTION") then
            ChatAssist_HistoryFrameScrollFrameString:SetText(ChatAssist:strjoin("\n", AuctionLog));
            ChatAssist_HistoryFrameScrollFrame:SetHeight(ChatAssist_HistoryFrameScrollFrameString:GetHeight()+10);
            ChatAssist_HistoryFrameScroll:UpdateScrollChildRect();
            ChatAssist_HistoryFrameKeywordButton:Enable();
            ChatAssist_HistoryFrameAuctionButton:Disable();
        end
    end
end

function ChatAssist:OpenChat(type)
    if (type == "SAY") then
        ChatFrame_OpenChat("/s ");
    elseif (type == "PARTY") then
        ChatFrame_OpenChat("/p ");
    elseif (type == "RAID") then
        ChatFrame_OpenChat("/ra ");
    elseif (type == "RAIDWARNING") then
        ChatFrame_OpenChat("/rw ");
    elseif (type == "BATTLEGROUND") then
        ChatFrame_OpenChat("/bg ");
    elseif (type == "GUILD") then
        ChatFrame_OpenChat("/g ");
    elseif (type == "OFFICIER") then
        ChatFrame_OpenChat("/o ");
    elseif (type == "WHISPER") then
        ChatFrame_OpenChat("/tell ");
    elseif (type == "TT") then
        ChatFrame_OpenChat("/tt ");
    elseif (type == "CHANNEL1") then
        ChatFrame_OpenChat("/1 ");
    elseif (type == "CHANNEL2") then
        ChatFrame_OpenChat("/2 ");
    elseif (type == "CHANNEL3") then
        ChatFrame_OpenChat("/3 ");
    elseif (type == "CHANNEL4") then
        ChatFrame_OpenChat("/4 ");
    elseif (type == "CHANNEL5") then
        ChatFrame_OpenChat("/5 ");
    elseif (type == "CHANNEL6") then
        ChatFrame_OpenChat("/6 ");
    elseif (type == "CHANNEL7") then
        ChatFrame_OpenChat("/7 ");
    elseif (type == "CHANNEL8") then
        ChatFrame_OpenChat("/8 ");
    elseif (type == "CHANNEL9") then
        ChatFrame_OpenChat("/9 ");
    elseif (type == "CHANNEL10") then
        ChatFrame_OpenChat("/10 ");
    end
end

function ChatAssist_TakeScreenshot(hideui)
    if (hideui == true and UIParent:IsVisible()) then
        CloseAllWindows();
        UIParent:Hide();
    end
    TakeScreenshot();
    if (hideui == true) then
        UIParent:Show();
    end
end

function ChatAssist:BoolToString(bool)
    if (bool) then
        return "on";
    else
        return "off";
    end
end

function ChatAssist:strjoin(delimiter, list)
    local len = table.getn(list);
    if (len == 0) then
        return "";
    end
    local string = list[1];
    for i = 2, len do 
        string = string .. delimiter .. list[i];
    end
    return string;
end

function ChatAssist:strsplit(delimiter, text)
    local list = {};
    local pos = 1;

    if (string.find("", delimiter, 1)) then
        table.insert(list, text);
        return list;
    end
    while 1 do
        local first, last = string.find(text, delimiter, pos);
        if first then
            table.insert(list, string.sub(text, pos, first-1));
            pos = last+1;
        else
            table.insert(list, string.sub(text, pos));
            break
        end
    end
    return list
end

function ChatAssist:isMultiByte(msg)
    local char = 0;

    for i=1, string.len(msg) do
        char = string.byte(string.sub(msg, i, i));
        if(char == nil or char > 128) then
            return true;
        end
    end
    return false;
end

function ChatAssist:OnMouseWheel(frame, value)
    if (IsShiftKeyDown()) then
        if (value > 0) then
            frame:ScrollToTop();
        else
            frame:ScrollToBottom();
        end
    elseif (IsControlKeyDown()) then
        if (value > 0) then
            frame:PageUp();
        else
            frame:PageDown();
        end
    else
        if (value > 0) then
            frame:ScrollUp();
        else
            frame:ScrollDown();
        end
    end
end

--
-- Command Functions
--

function ChatAssist:Command_toggleWindow(frame)
    self.db.profile.Window[frame] = not self.db.profile.Window[frame];
    self:Print("Window Toggle: " .. tostring(self.db.profile.Window[frame]));
end

function ChatAssist:Command_TellTarget(argstr)
    if (UnitExists("target")) then
        if (UnitIsPlayer("target") and UnitIsFriend("player", "target")) then
            SendChatMessage(argstr, "WHISPER", nil, UnitName("target"));
        end
    end
end

function ChatAssist:Command_addKeyword(argstr)
    -- Already exists?
    local exists = 0;
    for key,value in pairs(self.db.profile.Keywords) do
        if (argstr == value) then
            exists = 1;
        end
    end
    if (exists == 0) then
      table.insert(self.db.profile.Keywords, argstr);
      self:Print(format(L["Keyword (%s) added."], argstr));
    else
      self:Print(format(L["Keyword (%s) is already exists."], argstr));
    end
end

function ChatAssist:Command_delKeyword(argstr)
    local exists = 0;
    for key,value in pairs(self.db.profile.Keywords) do
        if (argstr == value) then
            table.remove(self.db.profile.Keywords, key);
            exists = 1;
        end
    end
    if (exists ~= 1) then
      self:Print(format(L["Keyword (%s) is not found."], argstr));
    else
      self:Print(format(L["Keyword (%s) deleted."], argstr));
    end
end

function ChatAssist:Command_listKeyword()
    self:Print(format(L["Keywords: %s"], ChatAssist:strjoin(",", self.db.profile.Keywords)));
end

function ChatAssist:Command_showChannelSticky()
    return self.db.profile.ChannelSticky;
end

function ChatAssist:Command_toggleChannelSticky()
    self.db.profile.ChannelSticky = not self.db.profile.ChannelSticky;
end

function ChatAssist:Command_showOfficerSticky()
    return self.db.profile.OfficerSticky;
end

function ChatAssist:Command_toggleOfficerSticky()
    self.db.profile.OfficerSticky = not self.db.profileOfficerSticky;
end

function ChatAssist:Command_showWhisperSticky()
    return self.db.profile.WhisperSticky;
end

function ChatAssist:Command_toggleWhisperSticky()
    self.db.profile.WhisperSticky = not self.db.profile.WhisperSticky;
end

function ChatAssist:Command_showTimestamp()
    return self.db.profile.Timestamp;
end

function ChatAssist:Command_toggleTimestamp()
    self.db.profile.Timestamp = not self.db.profile.Timestamp;
end

function ChatAssist:Command_getTimestampFormat()
    return self.db.profile.TimestampFormat;
end

function ChatAssist:Command_setTimestampFormat(argstr)
    self.db.profile.TimestampFormat = argstr;
end

function ChatAssist:Command_showLog()
    return self.db.profile.Logging;
end

function ChatAssist:Command_toggleLog()
    self.db.profile.Logging = not self.db.profile.Logging;
    LoggingChat(self.db.profile.Logging);
end

function ChatAssist:Command_History()
    if (not ChatAssist_HistoryFrame:IsVisible()) then
        ChatAssist_HistoryFrame:Show();
    end
    ChatAssist:HistoryFrameChange("KEYWORD");
end

function ChatAssist:Command_showColorName()
    return self.db.profile.ColorName;
end

function ChatAssist:Command_toggleColorName()
    self.db.profile.ColorName = not self.db.profile.ColorName;
    if (self.db.profile.ColorName) then
        ChatAssist:RebuildClassCache();
    end
end

function ChatAssist:Command_showShortTag()
    return self.db.profile.ShortTag;
end

function ChatAssist:Command_toggleShortTag()
    self.db.profile.ShortTag = not self.db.profile.ShortTag;
    if (self.db.profile.ShortTag) then
        self.db.profile.HideTag = false;
    end
end

function ChatAssist:Command_showHideTag()
    return self.db.profile.HideTag;
end

function ChatAssist:Command_toggleHideTag()
    self.db.profile.HideTag = not self.db.profile.HideTag;
    if (self.db.profile.HideTag) then
        self.db.profile.ShortTag = false;
    end
end

function ChatAssist:Command_showHideChannel()
    return self.db.profile.HideChannel;
end

function ChatAssist:Command_toggleHideChannel()
    self.db.profile.HideChannel = not self.db.profile.HideChannel;
end

function ChatAssist:Command_getUnicodeBlock(type)
    return self.db.profile.UnicodeBlock[type];
end

function ChatAssist:Command_toggleUnicodeBlock(argstr)
    self.db.profile.UnicodeBlock[argstr] = not self.db.profile.UnicodeBlock[argstr];
    self:Print("UnicodeBlock(" .. argstr .. "): " .. tostring(self.db.profile.UnicodeBlock[argstr]));
end

function ChatAssist:Command_getDefaultChat()
    return self.db.profile.DefaultChat;
end

function ChatAssist:Command_setDefaultChat(argstr)
    self.db.profile.DefaultChat = argstr;
end

function ChatAssist:Command_toggleOnScreen(argstr)
    local event = string.upper(argstr)
    self.db.profile.Event[event].OnScreen = not self.db.profile.Event[event].OnScreen;
end

function ChatAssist:Command_CCRebuild()
    if(self.db.profile.ColorName) then
        ChatAssist:RebuildClassCache();
    else
        self:Print(L["Error: ColorName is disabled."]);
    end
end

function ChatAssist:Command_Menu()
    local loaded = LoadAddOn('ChatAssist2Options');
    if (loaded == 1) then
        ShowUIPanel(ChatAssistOptionsFrame);
    end
end

function ChatAssist:Command_Frame()
    local frame = EnumerateFrames();
    while frame do
        if (frame:IsVisible() and MouseIsOver(frame)) then
            self:Print(frame:GetName());
        end
        frame = EnumerateFrames(frame);
    end
end

function ChatAssist:getProfile()
    return self.db.profile;
end

function ChatAssist:setProfile(profile)
    self.db.profile = profile;
end

function ChatAssist:getChatFont()
    return ChatFont;
end
